//-*-c++-*------------------------------------------------------------
// Pointer.cpp
//
// Author: Qi Yang (qiyang@alum.mit.edu)
//
// A testing example for class P<T> -- A template Smart Pointer class
// that performs reference counting garbage collection and supports
// subclassed pointers.
//
// Thanks: Stefan Tchekanov (stefant@iname.com). The code here copied
// peices from his SmartPtr class.
//
// HISTORY: 12-09-1999  Original version
//
// You may freely use or modify this code provided that above notice
// and notes on your changes are included in all derived versions.
//
//--------------------------------------------------------------------

#if defined(WIN32) || defined(_MSC_VER)
#include <windows.h>
#endif

#include <iostream>

#include "Pointer.h"

class B
{
public:
    B(const char* s_ = "") : _sb(s_) { }
    virtual void print(std::ostream& os = std::cout) {
        os << _sb << std::endl;
    }
protected:
    const char* _sb;
};

class A : public B
{
public:
    A(const char* s1_ = "", const char* s2_ = "") : _sa(s1_), B(s2_) { }
    void print(std::ostream& os = std::cout) {
        os << _sa << " "; B::print(os);
    }
protected:
    const char* _sa;
};

int main()
{
    // Create an object of derived class A.

    P<B> a = new A("Smart", "Pointer"); // a is the 1st reference
    P<B> b = a;                         // b is the 2nd reference
    P<A> c(a.Reference());              // c is the 3rd reference

    a = 0;                              // 1 reference reduced 
    c->print();                         // it calls A::print()

    // Create an object of the base class B

    P<B> x = new B("Trouble");  // x is the 1st reference

    // I would expect problem for the following code because 'x' (an
    // object of the base class B) is casted to pointer 'y' (class A
    // which is derived from B).

    P<A> y(x.Reference());              // y is the 2nd reference
    y->print();    // Strange enough, this does not crash
                   // but B::print() was called (MSVC 6.0).
                   // Any idea?
    return 0;
}
